import numpy as np

VERBOSE = False


class LinearRegressionGD:
    """Linear Regression Using Gradient Descent.
    Parameters
    ----------
    alpha : float
        Learning rate
    n_iterations : int
        No of passes over the training set
    Attributes
    ----------
    w_ : weights/ after fitting the model
    cost_ : total error of the model after each iteration
    """
    def __init__(self, alpha=0.1e-20, n_iterations=10000):
        self.alpha = alpha
        self.n_iterations = n_iterations

    def fit(self, x, y):
        """Fit the training data
        Parameters
        ----------
        x : array-like, shape = [n_samples, n_features]
            Training samples
        y : array-like, shape = [n_samples, n_target_values]
            Target values
        Returns
        -------
        self : object
        """

        self.cost_ = []
        self.w_ = np.zeros((x.shape[1], 1))
        m = x.shape[0]
        for _ in range(self.n_iterations):
            y_pred = np.dot(x, self.w_)
            residuals = y_pred - y
            gradient_vector = np.dot(x.T, residuals)
            self.w_ -= (self.alpha / m) * gradient_vector
            cost = np.sum((residuals**2)) / (2 * m)
            self.cost_.append(cost)

            if VERBOSE:
                print(self.w_)
        return self

    def predict(self, x):
        """ Predicts the value after the model has been trained.
        Parameters
        ----------
        x : array-like, shape = [n_samples, n_features]
            Test samples
        Returns
        -------
        Predicted value
        """
        return np.dot(x, self.w_)



import mock

from rComplexity.plotting.constants import pool_plotting_points_symbols
from rComplexity.plotting.plot_points import plot_points_from_file


def test_pool_plotting_points_symbols():
    assert next(pool_plotting_points_symbols) == "o"
    assert next(pool_plotting_points_symbols) == "x"
    assert next(pool_plotting_points_symbols) == "o"
    assert next(pool_plotting_points_symbols) == "x"
    assert next(pool_plotting_points_symbols) == "o"


@mock.patch("matplotlib.pyplot.show")
@mock.patch("matplotlib.pyplot.legend")
@mock.patch("matplotlib.pyplot.plot")
def test_plot_points_from_file(plt_mock_plot, plt_mock_legend, plt_mock_show):
    plot_points_from_file("rComplexity/plotting/tests/result_file_testing.txt",
                          5)

    assert plt_mock_plot.call_count == 5
    plt_mock_legend.assert_called_once_with()
    plt_mock_show.assert_called_once_with()



import random

import gym
import sys

if __name__ == "__main__":
    if len(sys.argv) != 3:
        print(
            "Usage: python3 play_vs_self.py <num_episodes> <num_steps_per_episode>"
        )
        sys.exit(-1)

    env = gym.make('ChessVsSelf-v0')

    #num_episodes = 100
    num_episodes = int(sys.argv[1])
    #num_steps_per_episode = 80
    num_steps_per_episode = int(sys.argv[2])

    collected_rewards = {1: [], -1: []}

    for i in range(num_episodes):
        initial_state = env.reset()
        # print('<'*5, '='*10, 'NEW GAME {}'.format(i+1), '='*10, '>'*5)
        # env._render()
        # print('<'*5, '-'*10, 'STARTING', '-'*10, '>'*5)

        player1 = 1
        player2 = -1
        total_rewards = {1: 0, -1: 0}
        done = False

        for j in range(num_steps_per_episode):
            if done or j == (num_steps_per_episode - 1):
                print('TOTAL GAME ', i, 'REWARD =', total_rewards)
                break

            for player in [player1, player2]:
                state = env.state
                board = state['board']
                kr_moves = state['kr_moves']
                captured = state['captured']

                moves = env.get_possible_moves(state, player)

                if len(moves) == 0:
                    a = env.resign_action()
                    print('<' * 5, '@' * 10, 'PLAYER RESIGNED', '@' * 10,
                          '>' * 5)
                else:
                    m = random.choice(moves)
                    a = env.move_to_actions(m)
                    # print('{:6s}'.format(env.convert_coords(m)), end=' ')

                # perform action
                state, reward, done, __ = env.step(a)
                total_rewards[player] += reward

        collected_rewards[1].append(total_rewards[1])
        collected_rewards[-1].append(total_rewards[-1])

    reward_1 = sum(collected_rewards[1])
    reward_2 = sum(collected_rewards[-1])

    print('\n')
    print('#' * 40)
    print('#' * 40)
    print('#' * 40)
    print("\nAVERAGE SCORE PLAYER 1: ", reward_1 / num_episodes)
    print("AVERAGE SCORE PLAYER 2: ", reward_2 / num_episodes)
    print('\n')



from WebMonitoring.API.resources.numerical_helpers import div_0


def get_results_resource_get_efficiency(
    result_args_get_time,
    result_args_old_get_time,
    list_sample_time,
    result_args_get_size,
    result_args_old_get_size,
    list_sample_size,
):
    result_args_get_efficiency = []
    result_args_old_get_efficiency = []

    for i in range(len(result_args_get_time)):
        result_args_get_efficiency.append(
            div_0(result_args_get_size[i], result_args_get_time[i]))

    for i in range(len(result_args_old_get_time)):
        result_args_old_get_efficiency.append(
            div_0(result_args_old_get_size[i], result_args_old_get_time[i]))

    list_sample_efficiency = []
    for i in range(len(list_sample_time)):
        list_sample_efficiency.append(
            div_0(list_sample_size[i], list_sample_time[i]))

    return (
        result_args_get_efficiency,
        result_args_old_get_efficiency,
        list_sample_efficiency,
    )



import io

import yaml


def remove_placement_contraints(dict):
    for service in dict["services"]:
        if dict["services"][service].get("deploy", None):
            dict["services"][service]["deploy"].pop("placement", None)
            if not dict["services"][service].get("deploy", None):
                dict["services"][service].pop("deploy", None)

    return dict


if __name__ == "__main__":
    with open("docker_config/docker-compose.yaml", "r") as stream:
        try:
            dict = yaml.safe_load(stream)

            dict_no_placement_contraints = remove_placement_contraints(dict)
            with io.open(
                    "docker_config/docker-compose-no-placement-constraints.yaml",
                    "w",
                    encoding="utf8",
            ) as outfile:
                yaml.dump(
                    dict_no_placement_contraints,
                    outfile,
                    default_flow_style=False,
                    allow_unicode=True,
                )

        except yaml.YAMLError as exc:
            print(exc)



from WebMonitoring.API.resources.numerical_helpers import lt_w_none, get_percent_w_none


def render_dict(
    result_args_get,
    result_args_old_get,
    list_sample,
    product_name,
    color,
    roundDecimal=2,
):

    return {
        "product": product_name,
        "total": {
            "monthly": round(result_args_get[3], roundDecimal),
            "weekly": round(result_args_get[2], roundDecimal),
            "daily": round(result_args_get[1], roundDecimal),
            "percent": {
                "value":
                get_percent_w_none(result_args_old_get[2], result_args_get[2]),
                "profit":
                lt_w_none(result_args_old_get[2], result_args_get[2]),
            },
        },
        "color": color,
        "lowest": {
            "monthly": {
                "value": round(result_args_get[6], roundDecimal),
                "profit": lt_w_none(result_args_old_get[6],
                                    result_args_get[6]),
            },
            "weekly": {
                "value": round(result_args_get[5], roundDecimal),
                "profit": lt_w_none(result_args_old_get[5],
                                    result_args_get[5]),
            },
            "daily": {
                "value": round(result_args_get[4], roundDecimal),
                "profit": lt_w_none(result_args_old_get[4],
                                    result_args_get[4]),
            },
        },
        "median": {
            "monthly": {
                "value": round(result_args_get[9], roundDecimal),
                "profit": lt_w_none(result_args_old_get[9],
                                    result_args_get[9]),
            },
            "weekly": {
                "value": round(result_args_get[8], roundDecimal),
                "profit": lt_w_none(result_args_old_get[8],
                                    result_args_get[8]),
            },
            "daily": {
                "value": round(result_args_get[7], roundDecimal),
                "profit": lt_w_none(result_args_old_get[7],
                                    result_args_get[7]),
            },
        },
        "highest": {
            "monthly": {
                "value": round(result_args_get[12], roundDecimal),
                "profit": lt_w_none(result_args_old_get[12],
                                    result_args_get[12]),
            },
            "weekly": {
                "value": round(result_args_get[11], roundDecimal),
                "profit": lt_w_none(result_args_old_get[11],
                                    result_args_get[11]),
            },
            "daily": {
                "value": round(result_args_get[10], roundDecimal),
                "profit": lt_w_none(result_args_old_get[10],
                                    result_args_get[10]),
            },
        },
        "samples": list_sample,
    }



from WebMonitoring.API.resources.last_seven_entries_metrics import (
    get_last_seven_entries_size, )


def get_results_resource_get_size(mysql, resource_name):
    result_args_get_size = []
    result_args_old_get_size = []
    list_sample = []

    try:
        conn = mysql.connect()
        cursor = conn.cursor()
        """
        Compatibility warning: PEP-249 specifies that any modified
        parameters must be returned. This is currently impossible
        as they are only available by storing them in a server
        variable and then retrieved by a query. Since stored
        procedures return zero or more result sets, there is no
        reliable way to get at OUT or INOUT parameters via callproc.
        The server variables are named @_procname_n, where procname
        is the parameter above and n is the position of the parameter
        (from zero). Once all result sets generated by the procedure
        have been fetched, you can issue a SELECT @_procname_0, ...
        query using .execute() to get any OUT or INOUT values.
        """
        args = [resource_name, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        cursor.callproc("resource_get_size", args)
        cursor.execute("SELECT @_resource_get_size_0, "
                       "@_resource_get_size_1, "
                       "@_resource_get_size_2, "
                       "@_resource_get_size_3, "
                       "@_resource_get_size_4, "
                       "@_resource_get_size_5, "
                       "@_resource_get_size_6, "
                       "@_resource_get_size_7, "
                       "@_resource_get_size_8, "
                       "@_resource_get_size_9, "
                       "@_resource_get_size_10, "
                       "@_resource_get_size_11, "
                       "@_resource_get_size_12 ")
        result_args_get_size = cursor.fetchone()

        cursor.callproc("resource_get_old_size", args)
        cursor.execute("SELECT @_resource_get_old_size_0, "
                       "@_resource_get_old_size_1, "
                       "@_resource_get_old_size_2, "
                       "@_resource_get_old_size_3, "
                       "@_resource_get_old_size_4, "
                       "@_resource_get_old_size_5, "
                       "@_resource_get_old_size_6, "
                       "@_resource_get_old_size_7, "
                       "@_resource_get_old_size_8, "
                       "@_resource_get_old_size_9, "
                       "@_resource_get_old_size_10, "
                       "@_resource_get_old_size_11, "
                       "@_resource_get_old_size_12 ")
        result_args_old_get_size = cursor.fetchone()

        list_sample = []
        samples = get_last_seven_entries_size(mysql, resource_name)
        for sample in samples:
            list_sample.append(sample["ResponseSize"])
    except Exception as e:
        print(e)
    finally:
        try:
            cursor.close()
            conn.close()
        except NameError:
            return (None, None, None)
        return (result_args_get_size, result_args_old_get_size, list_sample)



import difflib


def test_same_gitignore_dockerignore():
    """ Asserts that the versions of .gitignore and .dockerignore matches """
    with open(".gitignore", "r") as hosts0:
        with open(".dockerignore", "r") as hosts1:
            diff = difflib.unified_diff(
                hosts0.readlines(),
                hosts1.readlines(),
                fromfile="hosts0",
                tofile="hosts1",
            )
            for line in diff:
                assert line == None



from mysql_config.generate_init_sql import generate_init_sql
from mysql_config.tests.mock_sql import MOCK_SQL_INIT


def test_init_complete():
    generate_init_sql()
    with open("./mysql_config/init.sql", "r") as init_file:
        init_file_content = init_file.read()
        assert init_file_content == MOCK_SQL_INIT



import difflib

import yaml

from docker_config.generators.no_constraint_compose import remove_placement_contraints


def test_repo_docker_compose_no_placement_contraints():
    """ Asserts that the version of docker-compose-no-placement-constraints is the latest"""
    with open("docker_config/docker-compose-no-placement-constraints.yaml",
              "r") as hosts0:
        with open("docker_config/docker-compose.yaml", "r") as stream:
            dict = yaml.safe_load(stream)
            dict = remove_placement_contraints(dict)
            out_dict = yaml.dump(dict,
                                 default_flow_style=False,
                                 allow_unicode=True)

            diff = difflib.unified_diff(hosts0.readlines(),
                                        out_dict,
                                        fromfile="hosts0",
                                        tofile="hosts1")
            for line in diff:
                print(line)
                # assert line == None



import numpy as np
from sklearn.model_selection import train_test_split

from rafComputing.ML.features.feature_transformation import extract_features
from rafComputing.ML.features.feature_types import POLYNOMIAL_FEATURE_TYPE
from rafComputing.ML.helpers.generate_larger_evaluation_set import generate_larger_evaluation_set


def matrix_to_train_test(path,
                         feature_type=POLYNOMIAL_FEATURE_TYPE,
                         feature_val=1.0):
    """
        Loads a multiple-column space-separated matrix from @path,
        containing:
        Column 0: Metric key(e.g. input size)
        Column 1: Metric value(e.g. time)
        Returns as follows:
        Pair containing [ALL] extracted features and original read features (x, orig_x),
        Labels [ALL] y,
        Pair containing [Training] extracted features and original read features (X_train, orig_x_train),
        Pair containing [Test] extracted features and original read features (X_test, orig_x_test),
        Labels [Training] y_train,
        Labels [Test] y_test
    """
    matrix = np.loadtxt(path, usecols=range(2))

    # Parse Column 0: Metric key
    x = matrix[:, 0]
    # Parse Column 1: Metric value
    y = matrix[:, 1]

    # Ignore columns 2..
    x = x.reshape(-1, 1)
    y = y.reshape(-1, 1)

    # Store values for X before extract_features
    orig_x = x
    orig_x_train, orig_x_test, y_train, y_test = train_test_split(
        x, y, test_size=0.2, random_state=44)

    # Extract polynomial features
    x = extract_features(x, feature_type, feature_val)
    X_train = extract_features(orig_x_train, feature_type, feature_val)
    X_test = extract_features(orig_x_test, feature_type, feature_val)

    return (
        (x, orig_x),
        y,
        (X_train, orig_x_train),
        (X_test, orig_x_test),
        y_train,
        y_test,
    )


def matrix_to_train_test_w_generate_larger_evaluation_set(
    path,
    feature_type=POLYNOMIAL_FEATURE_TYPE,
    feature_val=1.0,
    custom_generate_larger_evaluation_set=generate_larger_evaluation_set):
    """
        Same as matrix_to_train_test but returns two test sets:
        1. Same Test set as matrix_to_train_test
        2. Unlabeled Test set:
            (X_test2, orig_x_test2),
    """
    (x, orig_x), y, (X_train, orig_x_train), (
        X_test, orig_x_test), y_train, y_test = matrix_to_train_test(
            path, feature_type=feature_type, feature_val=feature_val)

    orig_x_test2 = custom_generate_larger_evaluation_set(orig_x)
    X_test2 = extract_features(orig_x_test2, feature_type, feature_val)

    return (
        (x, orig_x),
        y,
        (X_train, orig_x_train),
        (X_test, orig_x_test),
        y_train,
        y_test,
        (X_test2, orig_x_test2),
    )
